package com.guildasoft.bluechatroom2.adapter

import android.bluetooth.BluetoothAdapter
import android.bluetooth.BluetoothDevice
import android.bluetooth.BluetoothServerSocket
import android.bluetooth.BluetoothSocket
import android.content.Context
import android.os.Bundle
import android.os.Handler
import android.util.Log
import com.guildasoft.bluechatroom2.entity.AppMessage
import com.guildasoft.bluechatroom2.entity.Constants
import com.guildasoft.bluechatroom2.utils.MessageUtils
import com.guildasoft.bluechatroom2.utils.MessageUtilsImpl
import java.io.IOException
import java.io.InputStream
import java.io.OutputStream
import java.util.*

class BluetoothChatService(context: Context, handler: Handler){
    private var conversor: MessageUtils = MessageUtilsImpl()
    private var mAdapter: BluetoothAdapter? = null
    private var mHandler: Handler? = null
    private var mSecureAcceptThread: AcceptThread? = null
    private var mInsecureAcceptThread: AcceptThread? = null
    private var mConnectThread: ConnectThread? = null
    private var mConnectedThread: ConnectedThread? = null
    private var mState: Int = 0
    private var mNewState: Int = 0

    private val TAG: String = javaClass.simpleName
    private val MY_UUID_SECURE = UUID.fromString("3ad74fd6-efd9-482e-b38f-73d8dbb382e1")
    private val MY_UUID_INSECURE = UUID.fromString("d620cd2b-e0a4-435b-b02e-40324d57195b")
    private val NAME_SECURE = "BluetoothChatSecure"
    private val NAME_INSECURE = "BluetoothChatInsecure"

    companion object {
        val STATE_NONE = 0
        val STATE_LISTEN = 1
        val STATE_CONNECTING = 2
        val STATE_CONNECTED = 3
        val NEW_INCOMING_CONNECTION = 8
    }

    init {
        mAdapter = BluetoothAdapter.getDefaultAdapter()
        mState = STATE_NONE
        mNewState = mState
        mHandler = handler
    }

    @Synchronized fun start() {
        Log.d(TAG, "start")

        if (mConnectThread != null) {
            mConnectThread?.cancel()
            mConnectThread = null
        }

        if (mConnectedThread != null) {
            mConnectedThread?.cancel()
            mConnectedThread = null
        }
        if (mSecureAcceptThread == null) {
            mSecureAcceptThread = AcceptThread(true)
            mSecureAcceptThread?.start()
        }
        if (mInsecureAcceptThread == null) {
            mInsecureAcceptThread = AcceptThread(false)
            mInsecureAcceptThread?.start()
        }
    }

    @Synchronized fun connect(device: BluetoothDevice?, secure: Boolean) {

        Log.d(TAG, "connect to: " + device)

        if (mState == STATE_CONNECTING) {
            if (mConnectThread != null) {
                mConnectThread?.cancel()
                mConnectThread = null
            }
        }

        if (mConnectedThread != null) {
            mConnectedThread?.cancel()
            mConnectedThread = null
        }

        if(BluetoothAdapter.getDefaultAdapter().bondedDevices.contains(device)){
            mConnectThread = ConnectThread(device, secure)
            mConnectThread?.start()
        }
    }

    @Synchronized fun connected(socket: BluetoothSocket?, device: BluetoothDevice?, socketType: String) {
        Log.d(TAG, "connected, Socket Type:" + socketType)

        if (mConnectThread != null) {
            mConnectThread?.cancel()
            mConnectThread = null
        }

        if (mConnectedThread != null) {
            mConnectedThread?.cancel()
            mConnectedThread = null
        }

        mConnectedThread = ConnectedThread(socket, socketType)
        mConnectedThread?.start()

        val msg = mHandler?.obtainMessage(Constants.MESSAGE_DEVICE_NAME)
        val bundle = Bundle()
        bundle.putString(Constants.DEVICE_NAME, device?.name)
        msg?.data = bundle
        mHandler?.sendMessage(msg!!)
    }

    @Synchronized fun stop() {
        Log.d(TAG, "stop")

        if (mConnectThread != null) {
            mConnectThread?.cancel()
            mConnectThread = null
        }

        if (mConnectedThread != null) {
            mConnectedThread?.cancel()
            mConnectedThread = null
        }

        if (mSecureAcceptThread != null) {
            mSecureAcceptThread?.cancel()
            mSecureAcceptThread = null
        }

        if (mInsecureAcceptThread != null) {
            mInsecureAcceptThread?.cancel()
            mInsecureAcceptThread = null
        }
        mState = STATE_NONE
    }

    fun write(out: ByteArray) {
        var r: ConnectedThread?
        synchronized(this) {
            if (mState != STATE_CONNECTED) return
            r = mConnectedThread
        }
        r?.write(out)
    }

    fun sendMessage(appMessage: AppMessage){
        val mountedMessage = conversor.mountCompleteMessage(
            appMessage.messageType,
            appMessage.uuid,
            appMessage.owner,
            appMessage.text
        )
        write(mountedMessage.toByteArray())
    }

    private fun connectionFailed() {
        val msg = mHandler?.obtainMessage(Constants.MESSAGE_TOAST)
        val bundle = Bundle()
        bundle.putString(Constants.TOAST, "Erro na conexão!")
        msg?.data = bundle
        mHandler?.sendMessage(msg!!)

        mState = STATE_NONE

        this@BluetoothChatService.start()
    }

    private fun connectionLost(goneDeviceName : String) {
        val msg = mHandler?.obtainMessage(Constants.MESSAGE_CONNECTION_LOST)
        val bundle = Bundle()
        bundle.putString(Constants.CONNECTION_LOST, goneDeviceName)
        msg?.data = bundle
        mHandler?.sendMessage(msg!!)

        mState = STATE_NONE

        this@BluetoothChatService.start()
    }

    private inner class AcceptThread(secure: Boolean) : Thread() {
        private val mmServerSocket: BluetoothServerSocket?
        private val mSocketType: String
        private val security: Boolean

        init {
            var tmp: BluetoothServerSocket? = null
            security = secure
            mSocketType = if (secure) "Secure" else "Insecure"

            try {
                if (secure) {
                    tmp = mAdapter?.listenUsingRfcommWithServiceRecord(NAME_SECURE,
                            MY_UUID_SECURE)
                } else {
                    tmp = mAdapter?.listenUsingInsecureRfcommWithServiceRecord(
                            NAME_INSECURE, MY_UUID_INSECURE)
                }
            } catch (e: IOException) {
                Log.e(TAG, "Socket Type: " + mSocketType + "listen() failed", e)
            }

            mmServerSocket = tmp
            mState = STATE_LISTEN
        }

        override fun run() {

            Log.d(TAG, "Socket Type: " + mSocketType +
                    "BEGIN mAcceptThread" + this)
            name = "AcceptThread" + mSocketType

            var socket: BluetoothSocket?

            while (mState != STATE_CONNECTED) {
                try {
                    socket = mmServerSocket?.accept(1000)
                } catch (e: IOException) {
                    socket = null
                }

                if (socket != null) {
                    synchronized(this@BluetoothChatService) {
                        when (mState) {
                            STATE_LISTEN, STATE_CONNECTING -> {
                                mHandler?.obtainMessage(Constants.MESSAGE_STATE_CHANGE, NEW_INCOMING_CONNECTION, if (security) 1 else 0, socket)?.sendToTarget()
                            }
                            STATE_NONE, STATE_CONNECTED ->
                                try {
                                    socket.close()
                                } catch (e: IOException) {
                                    Log.e(TAG, "Could not close unwanted socket", e)
                                }
                            else -> {
                            }
                        }
                    }
                }
            }
            Log.i(TAG, "END mAcceptThread, socket Type: " + mSocketType)
        }

        fun cancel() {
            Log.d(TAG, "Socket Type" + mSocketType + "cancel " + this)
            try {
                mmServerSocket?.close()
            } catch (e: IOException) {
                Log.e(TAG, "Socket Type" + mSocketType + "close() of server failed", e)
            }
        }
    }
    private inner class ConnectThread(private val mmDevice: BluetoothDevice?, secure: Boolean) : Thread() {
        private val mmSocket: BluetoothSocket?
        private val mSocketType: String

        init {
            var tmp: BluetoothSocket? = null
            mSocketType = if (secure) "Secure" else "Insecure"
            try {
                if(secure) {
                    tmp = mmDevice?.createRfcommSocketToServiceRecord(
                            MY_UUID_SECURE)
                } else {
                    tmp = mmDevice?.createInsecureRfcommSocketToServiceRecord(
                            MY_UUID_INSECURE)
                }
            } catch (e: IOException) {
                Log.e(TAG, "Socket Type: " + mSocketType + "create() failed", e)
            }

            mmSocket = tmp
            mState = STATE_CONNECTING
        }

        override fun run() {

            Log.i(TAG, "BEGIN mConnectThread SocketType:" + mSocketType)
            name = "ConnectThread" + mSocketType

            mAdapter?.cancelDiscovery()

            try {
                mmSocket?.connect()
            } catch (e: IOException) {
                try {
                    mmSocket?.close()
                } catch (e2: IOException) {
                    Log.e(TAG, "unable to close() " + mSocketType +
                            " socket during connection failure", e2)
                }
                connectionFailed()
                return
            }
            synchronized(this@BluetoothChatService) {
                mConnectThread = null
            }
            connected(mmSocket, mmDevice, mSocketType)
        }

        fun cancel() {
            try {
                mmSocket?.close()
            } catch (e: IOException) {
                Log.e(TAG, "close() of connect $mSocketType socket failed", e)
            }

        }
    }

    private inner class ConnectedThread(private val mmSocket: BluetoothSocket?, socketType: String) : Thread() {

        private val mmInStream: InputStream?
        private val mmOutStream: OutputStream?

        init {
            Log.d(TAG, "create ConnectedThread: " + socketType)
            var tmpIn: InputStream? = null
            var tmpOut: OutputStream? = null

            try {
                tmpIn = mmSocket?.inputStream
                tmpOut = mmSocket?.outputStream
            } catch (e: IOException) {
                Log.e(TAG, "Os sockets temporarios nao foram criados", e)
            }
            mmInStream = tmpIn
            mmOutStream = tmpOut
            mState = STATE_CONNECTED

            mHandler?.obtainMessage(Constants.MESSAGE_STATE_CHANGE, mState, -1, mmSocket?.remoteDevice)?.sendToTarget()
        }

        override fun run() {
            var buffer = ByteArray(1024)
            var bytes: Int

            while (mState == STATE_CONNECTED) {
                try {
                    bytes = mmInStream?.read(buffer) ?: 0
                    val messageObject: AppMessage = conversor.convertMessageBytesToObject(buffer, bytes)
                    mHandler?.obtainMessage(Constants.MESSAGE_READ, -1, -1, messageObject)?.sendToTarget()
                    buffer = ByteArray(1024)
                } catch (e: IOException) {
                    Log.e(TAG, "disconnected", e)
                    connectionLost(mmSocket?.remoteDevice?.name!!)
                    break
                }
            }
        }

        fun write(buffer: ByteArray) {
            try {
                mmOutStream?.write(buffer)
                mHandler?.obtainMessage(Constants.MESSAGE_WRITE, -1, -1, buffer)
                        ?.sendToTarget()
            } catch (e: IOException) {
                Log.e(TAG, "Exception during write", e)
            }
        }

        fun cancel() {
            try {
                mmSocket?.close()
            } catch (e: IOException) {
                Log.e(TAG, "close() of connect socket failed", e)
            }
        }
    }
}
